package icecast

import (
	"context"
	"encoding/xml"
	"fmt"
	"io"
	"log"
	"net/http"
	"strconv"
	"time"

	"git.autistici.org/ale/autoradio"
	pb "git.autistici.org/ale/autoradio/proto"
)

var (
	icecastStatusPage           = "/status-autoradio.xsl"
	icecastStatusUpdateInterval = 2 * time.Second
)

// Icecast returns empty fields in our status handler, which we'll
// need to turn into integers (the xml unmarshaler will return an
// error in this specific case), so we use a separate type for
// decoding the status page output. This would be much simpler if I
// knew how to get the XSLT to put a default value in the output
// instead of an empty field...
type icecastMountStatus struct {
	Name         string `xml:"name,attr"`
	Listeners    string `xml:"listeners"`
	BitRate      string `xml:"bitrate"`
	Quality      string `xml:"quality"`
	VideoQuality string `xml:"video-quality"`
	FrameSize    string `xml:"frame-size"`
	FrameRate    string `xml:"frame-rate"`
}

type icecastStatus struct {
	XMLName xml.Name             `xml:"status"`
	Mounts  []icecastMountStatus `xml:"mount"`
}

func parseIcecastStatus(r io.Reader) (*icecastStatus, error) {
	var doc icecastStatus
	if err := xml.NewDecoder(r).Decode(&doc); err != nil {
		return nil, err
	}
	return &doc, nil
}

func fetchIcecastStatus(ctx context.Context, port int) (*icecastStatus, error) {
	req, err := http.NewRequest("GET", fmt.Sprintf("http://localhost:%d%s", port, icecastStatusPage), nil)
	if err != nil {
		return nil, err
	}
	resp, err := httpClient.Do(req.WithContext(ctx))
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("HTTP status %d", resp.StatusCode)
	}
	return parseIcecastStatus(resp.Body)
}

func convertIcecastStatus(status *icecastStatus) []*pb.IcecastMount {
	out := make([]*pb.IcecastMount, 0, len(status.Mounts))
	for _, m := range status.Mounts {
		outm := pb.IcecastMount{
			Path:         autoradio.IcecastPathToMountPath(m.Name),
			Listeners:    toi(m.Listeners),
			BitRate:      toi(m.BitRate),
			Quality:      tof(m.Quality),
			VideoQuality: tof(m.VideoQuality),
			FrameSize:    m.FrameSize,
			FrameRate:    tof(m.FrameRate),
		}
		out = append(out, &outm)
	}
	return out
}

func toi(s string) int32 {
	if i, err := strconv.Atoi(s); err == nil {
		return int32(i)
	}
	return 0
}

func tof(s string) float32 {
	if f, err := strconv.ParseFloat(s, 32); err == nil {
		return float32(f)
	}
	return 0
}

func (c *Controller) statusUpdater(ctx context.Context) {
	tick := time.NewTicker(icecastStatusUpdateInterval)
	defer tick.Stop()
	for {
		select {
		case <-tick.C:
			uctx, cancel := context.WithTimeout(ctx, icecastHTTPTimeout)
			status, err := fetchIcecastStatus(uctx, autoradio.IcecastPort)
			cancel()

			c.statusMx.Lock()
			if err != nil {
				log.Printf("icecast_status: error: %v", err)
				c.icecastOk = false
				c.mounts = nil
			} else {
				c.icecastOk = true
				c.mounts = convertIcecastStatus(status)
			}
			c.statusMx.Unlock()

		case <-ctx.Done():
			return
		}
	}
}

func (c *Controller) GetStatus() ([]*pb.IcecastMount, bool) {
	c.statusMx.Lock()
	defer c.statusMx.Unlock()
	return c.mounts, c.icecastOk
}
